from __future__ import absolute_import
input_name = '../examples/linear_elasticity/linear_elastic_damping.py'
output_name_trunk = 'test_linear_elastic_damping'

from tests_basic import TestInputEvolutionary

class Test( TestInputEvolutionary ):
    pass
