from __future__ import absolute_import
input_name = '../examples/linear_elasticity/elastic_contact_planes.py'
output_name = 'test_elastic_contact_planes.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
