from __future__ import absolute_import
input_name = '../examples/acoustics/vibro_acoustic3d.py'
output_name = 'test_vibro_acoustic3d'

from tests_basic import TestInput
class Test(TestInput):
    pass
