# c: 20.03.2008, r: 20.03.2008
from __future__ import absolute_import
input_name = '../examples/diffusion/sinbc.py'
output_name = 'test_sinbc.vtk'

from tests_basic import TestInput
class Test( TestInput ):
    pass
